#include <ecl/ecl.h>

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

static int boot_done = 0;

static void
create_lisp_on_sv(SV *rsv, char *pkg_name)
{
    char classname[200];
    SV *sv = SvRV(rsv);
    HV *phv;
    if (strchr(pkg_name,':')!=0)
	strcpy(classname,pkg_name); /* if full name given, use it */
    else
	sprintf(classname, "Language::Lisp::ECLs::%s::",pkg_name); /*otherwise ...*/
    phv = get_hv(classname,TRUE);
    sv_bless(rsv,phv);
}

static SV*
create_lisp_sv(char *pkg_name, cl_object obj)
{
    SV *sv = newSVpv((char*)&obj,4);
    SV *rsv = newRV_noinc(sv); /* after this sv will have refcnt 1 (fortunately) */
    create_lisp_on_sv(rsv, pkg_name);
    return rsv;
}

/* given blessed reference, which is actually pointer to cl_object,
 * return this as lisp object */
static cl_object
sv2cl_nocheck(SV *sv)
{
    cl_object clo;
    /* proper checks are not within scope of this fun */
    SV *sv_deref = SvRV(sv);
    memcpy(&clo,SvPV_nolen(sv_deref),4);
    return clo;
}
static cl_object
sv2cl(SV *sv)
{
    if (sv_isobject(sv) && sv_derived_from(sv, "Language::Lisp::ECLs")) {
	cl_object clo;
	SV *sv_deref = SvRV(sv);
	memcpy(&clo,SvPV_nolen(sv_deref),4);
	return clo;
    } else {
        croak("sv2cl: passed not a subclass of Language::Lispp::ECLs");
    }
}

static void
free_cl(SV *sv)
{
    cl_object o = sv2cl(sv);
    if (type_of(o) == t_base_string) {
	fprintf(stderr,"freeing a base_string\n");
        GC_free(o->base_string.self);
    } else {
	fprintf(stderr,"free of type %d not impl!\n",type_of(o));
    }
}

MODULE = Language::Lisp::ECLs::Package		PACKAGE = Language::Lisp::ECLs::Package		

SV *
stringify(clsv)
        SV *clsv
    PREINIT:
        cl_object clo, np;
    CODE:
        clo = sv2cl(clsv);
	if (type_of(clo) == t_package) {
	    np = cl_package_name(clo);
	    RETVAL = newSVpvf("package n=%s",
		(type_of(np)==t_base_string?np->base_string.self:"??")
	      );
	} else {
	    croak("can not stringify non-t_package within ...::Package package");
	}
    OUTPUT:
    	RETVAL

MODULE = Language::Lisp::ECLs::Symbol		PACKAGE = Language::Lisp::ECLs::Symbol		

SV *
stringify(clsv)
        SV *clsv
    PREINIT:
        cl_object clo, n, p, np;
    CODE:
        clo = sv2cl(clsv);
	if (type_of(clo) == t_symbol) {
	    n = ecl_symbol_name(clo);
	    p = ecl_symbol_package(clo);
	    np = cl_package_name(p);
	    RETVAL = newSVpvf("symbol n=%s p=%s",
		(type_of(n)==t_base_string?n->base_string.self:"??"),
		(type_of(np)==t_base_string?np->base_string.self:"??")
	      );
	} else {
	    croak("can not stringify non-t_symbol within ...::Symbol package");
	}
    OUTPUT:
    	RETVAL

MODULE = Language::Lisp::ECLs::String		PACKAGE = Language::Lisp::ECLs::String		

SV *
stringify(clsv)
        SV *clsv
    PREINIT:
        cl_object clo, n, p, np;
    CODE:
        clo = sv2cl(clsv);
	if (type_of(clo) == t_base_string) {
	    RETVAL = newSVpv(clo->base_string.self,clo->base_string.fillp);
	} else {
	    croak("can not stringify non-t_base_string within ...::String package");
	}
    OUTPUT:
    	RETVAL

MODULE = Language::Lisp::ECLs::Generic		PACKAGE = Language::Lisp::ECLs::Generic		

SV *
stringify(clsv)
        SV *clsv
    PREINIT:
	char *types[] = {
	"t_start",
	"t_list",
	"t_character",
	"t_fixnum",
#ifdef ECL_SHORT_FLOAT
	"t_shortfloat",
#endif
	"t_bignum",
	"t_ratio",
	"t_singlefloat",
	"t_doublefloat",
#ifdef ECL_LONG_FLOAT
	"t_longfloat",
#endif
	"t_complex",
	"t_symbol",
	"t_package",
	"t_hashtable",
	"t_array",
	"t_vector",
#ifdef ECL_UNICODE
	"t_string",
#endif
	"t_base_string",
	"t_bitvector",
	"t_stream",
	"t_random",
	"t_readtable",
	"t_pathname",
	"t_bytecodes",
	"t_cfun",
	"t_cclosure",
#ifdef CLOS
	"t_instance",
#else
	"t_structure",
#endif /* CLOS */
#ifdef ECL_THREADS
	"t_process",
	"t_lock",
	"t_condition_variable",
#endif
	"t_codeblock",
	"t_foreign",
	"t_frame",
	"t_end",
	"t_other",
	"t_contiguous"
	    /*FREE = 127 */      /*  free object  */
	};
	int t;
	char *h;
	cl_object o = sv2cl(clsv);
    CODE:
	t = type_of(o);
	if (t==127) {h="FREE";}
	else {h=types[t];}
	fprintf(stderr,"can not stringify %s within ...::Generic package ... yet\n", h);
	RETVAL = "?";
    OUTPUT:
    	RETVAL

MODULE = Language::Lisp::ECLs		PACKAGE = Language::Lisp::ECLs		

int
cl_boot()
    PREINIT:
        char *argv1[] = {""};
    CODE:
	//argc, argv TODO int argc
	//argc, argv TODO char **argv
        RETVAL = cl_boot(0,argv1);
	boot_done = 1;
    OUTPUT:
    	RETVAL

void
cl_shutdown()
    CODE:
        cl_shutdown();

SV *
_eval_string(s)
	char *s
    PREINIT:
	cl_object def;
	cl_object n, p;
	cl_object res;
    CODE:
	//if (!boot_done)
	//    XS_Language__Lisp__ECL_cl_boot(aTHX);
        def = c_string_to_object(s);
	res = cl_eval(def);
	switch (type_of(res)) {
	case t_fixnum:
	    RETVAL = newSViv(fix(res));
	    break;
	case t_base_string:
	    RETVAL = newSVpv(res->base_string.self,res->base_string.fillp);
	    break;
#ifdef ECL_UNICODE
	case t_string:
	    fprintf(stderr,"wtf - t_string(%d)\n",res->string.fillp);
	    RETVAL = newSVpv(res->string.self,res->string.fillp);
	    break;
#endif
	case t_package:
	    /*
	    n = cl_package_name(res);
	    fprintf(stderr,"n-type %d, n=%s\n",
	        type_of(n),
		(type_of(n)==t_base_string?n->base_string.self:"??"));
		*/
	    RETVAL = create_lisp_sv("Package", res);
	    break;
	case t_symbol:
		/*
	    n = ecl_symbol_name(res);
	    p = ecl_symbol_package(res);
	    fprintf(stderr,"n-type %d, n=%s p-type %d\n",
	        type_of(n),
		(type_of(n)==t_base_string?n->base_string.self:"??"),
		type_of(p));
		*/
	    RETVAL = create_lisp_sv("Symbol", res);
	    break;
	default:
	    fprintf(stderr,"type %d not impl!\n",type_of(res));
	    RETVAL = create_lisp_sv("Generic", res);
	    break;
	}
    OUTPUT:
    	RETVAL


SV *
_eval(lispobj)
	SV *lispobj
    PREINIT:
	cl_object def = sv2cl(lispobj);
	cl_object res;
    CODE:
	//if (!boot_done)
	//    XS_Language__Lisp__ECL_cl_boot(aTHX);
	res = cl_eval(def);
	switch (type_of(res)) {
	case t_fixnum:
	    RETVAL = newSViv(fix(res));
	    break;
	case t_base_string:
	    RETVAL = newSVpv(res->base_string.self,res->base_string.fillp);
	    break;
	case t_package:
	    RETVAL = create_lisp_sv("Package", res);
	    break;
	case t_symbol:
	    RETVAL = create_lisp_sv("Symbol", res);
	    break;
	default:
	    fprintf(stderr,"type %d not impl!\n",type_of(res));
	    RETVAL = create_lisp_sv("Generic", res);
	    break;
	}
    OUTPUT:
    	RETVAL


SV *
create_string(sv)
	SV *sv
    PREINIT:
        int len;
	char *str;
        cl_object x;
    CODE:
        str = SvPV(sv,len);
        x = cl_alloc_simple_base_string(len);
        memcpy(x->base_string.self, str, len);
	RETVAL = create_lisp_sv("String",x);
    OUTPUT:
        RETVAL

